import json
import logging

from django import forms
from django.contrib import messages
from django.core.validators import validate_email
from django.utils.translation import ugettext_lazy as _

from .models import EmailTemplate, EmailTemplateQuerySet

logger = logging.getLogger(__name__)


class JSONWidget(forms.Textarea):

    """Pretty print JSON in a text area."""

    DEFAULT_ATTRS = {'class': 'vLargeTextField', 'rows': 15}

    def format_value(self, value):
        """Pretty format JSON text."""
        value = value or '{}'
        if not isinstance(value, str):
            raise TypeError("Value must JSON parseable string instance")
        value = json.loads(value)
        return json.dumps(value, indent=4, sort_keys=True)

    def render(self, name, value, attrs=DEFAULT_ATTRS):
        value = self.format_value(value)
        return super(JSONWidget, self).render(name, value, attrs=attrs)


class MultiEmailField(forms.Field):

    """Taken from https://docs.djangoproject.com/en/1.11/ref/forms/validation/#form-field-default-cleaning """  # noqa

    def to_python(self, value):
        """Normalize data to a list of strings."""
        if isinstance(value, list):
            return value

        if not value:
            return []

        return [v.strip() for v in value.split(',')]

    def validate(self, value):
        """Check if value consists only of valid emails."""
        # Use the parent's handling of required fields, etc.
        super(MultiEmailField, self).validate(value)
        for email in value:
            validate_email(email)


class MultiEmailTemplateField(forms.Field):

    """Convert comma-separated ids into EmailTemplate queryset."""

    def to_python(self, value):
        """Normalize data to a queryset of EmailTemplates."""
        if isinstance(value, EmailTemplateQuerySet):
            return value

        if not value:
            return EmailTemplate.objects.none()

        values = [int(i) for i in value.split(',')]
        return EmailTemplate.objects.filter(pk__in=values)


class EmailTestForm(forms.Form):

    """Renders email template on intermediate page."""

    from_email = forms.EmailField(
        label=_("From"),
        help_text=_("Email address to be used as the sender")
    )
    reply_to = MultiEmailField(
        label=_("Reply-To"),
        help_text=_("Comma separated list of email addresses")
    )
    to = MultiEmailField(
        label=_("To"),
        help_text=_("Comma separated list of email addresses")
    )
    cc = MultiEmailField(
        label=_("cc"),
        help_text=_("Comma separated list of email addresses"),
        required=False
    )
    bcc = MultiEmailField(
        label=_("bcc"),
        help_text=_("Comma separated list of email addresses"),
        required=False
    )
    context = forms.CharField(
        widget=forms.Textarea,
        required=False,
        help_text=_("JSON used to render the subject and body templates")
    )
    # comma separated list of template ids.
    templates = MultiEmailTemplateField(
        widget=forms.HiddenInput()
    )

    def clean_context(self):
        """Load text input back into JSON."""
        context = self.cleaned_data['context'] or '{}'
        try:
            return json.loads(context)
        except (TypeError, ValueError) as ex:
            raise forms.ValidationError(_("Invalid JSON: %s" % ex))

    def _create_message(self, template):
        """Create EmailMultiMessage from form data."""
        return template.create_message(
            self.cleaned_data['context'],
            from_email=self.cleaned_data['from_email'],
            to=self.cleaned_data['to'],
            cc=self.cleaned_data['cc'],
            bcc=self.cleaned_data['bcc']
        )

    def send_emails(self, request):
        """Send test emails."""
        for template in self.cleaned_data.get('templates'):
            email = self._create_message(template)
            try:
                email.send()
            except Exception as ex:
                logger.exception("Error sending test email")
                messages.error(
                    request,
                    _("Error sending test email '%s': %s" % (template.name, ex))
                )
            else:
                messages.success(
                    request,
                    _("'%s' email sent to '%s'" % (template.name, ', '.join(email.to)))
                )



from unittest import mock

from django.conf import settings
from django.core.exceptions import ValidationError
from django.core.mail import EmailMultiAlternatives
from django.template import TemplateDoesNotExist, TemplateSyntaxError
from django.test import TestCase

from ..models import EmailTemplate


class EmailTemplateQuerySetTests(TestCase):

    """appmail.models.EmailTemplateQuerySet model tests."""

    def test_active(self):
        template1 = EmailTemplate(name='test1', language='en-us').save()
        template2 = EmailTemplate(name='test2', language='en-us', is_active=False).save()
        self.assertEqual(EmailTemplate.objects.active().get(), template1)

    def test_current(self):
        # manually setting the version in the wrong order, so the first
        # template is actually the last, when ordered by version.
        template1 = EmailTemplate(name='test', language='en-us', version=1).save()
        template2 = EmailTemplate(name='test', language='en-us', version=0).save()
        self.assertEqual(EmailTemplate.objects.current('test'), template1)
        self.assertEqual(EmailTemplate.objects.current('test', language='klingon'), None)

    def test_version(self):
        template1 = EmailTemplate(name='test', language='en-us', version=1).save()
        template2 = EmailTemplate(name='test', language='en-us', version=0).save()
        self.assertEqual(EmailTemplate.objects.version('test', 1), template1)
        self.assertEqual(EmailTemplate.objects.version('test', 0), template2)


class EmailTemplateTests(TestCase):

    """appmail.models.EmailTemplate model tests."""

    def test_defaults(self):
        template = EmailTemplate()
        self.assertEqual(template.language, settings.LANGUAGE_CODE)
        self.assertEqual(template.from_email, settings.DEFAULT_FROM_EMAIL)
        self.assertEqual(template.reply_to, settings.DEFAULT_FROM_EMAIL)
        self.assertEqual(template.reply_to_list, [settings.DEFAULT_FROM_EMAIL])
        self.assertEqual(template.version, 0)

        template.reply_to = 'fred@example.com, ginger@example.com'
        self.assertEqual(template.reply_to_list, ['fred@example.com', 'ginger@example.com'])

    @mock.patch.object(EmailTemplate, 'clean')
    def test_save(self, mock_clean):
        template = EmailTemplate(
            subject='test ßmail',
            body_text='this is plain text',
            body_html='this is <b>html</b>'
        )
        with mock.patch('appmail.models.VALIDATE_ON_SAVE', False):
            template.save()
            self.assertEqual(mock_clean.call_count, 0)
        with mock.patch('appmail.models.VALIDATE_ON_SAVE', True):
            template.save()
            self.assertEqual(mock_clean.call_count, 1)
            # test the override
            template.save(validate=False)
            self.assertEqual(mock_clean.call_count, 1)

    @mock.patch.object(EmailTemplate, 'render_subject')
    def test__validate_subject(self, mock_render):
        template = EmailTemplate()
        mock_render.side_effect = TemplateDoesNotExist('foo.html')
        self.assertEqual(
            template._validate_subject(),
            {'subject': "Template does not exist: foo.html"}
        )
        mock_render.side_effect = TemplateSyntaxError('No can do')
        self.assertEqual(
            template._validate_subject(),
            {'subject': "No can do"}
        )
        mock_render.side_effect = None
        self.assertEqual(
            template._validate_subject(),
            {}
        )
        mock_render.side_effect = Exception("Something else")
        self.assertRaises(Exception, template._validate_subject)

    @mock.patch.object(EmailTemplate, 'render_body')
    def test__validate_body(self, mock_render):
        template = EmailTemplate()
        mock_render.side_effect = TemplateDoesNotExist('foo.html')
        self.assertEqual(
            template._validate_body(content_type=EmailTemplate.CONTENT_TYPE_PLAIN),
            {'body_text': "Template does not exist: foo.html"}
        )
        mock_render.side_effect = TemplateSyntaxError('No can do')
        self.assertEqual(
            template._validate_body(content_type=EmailTemplate.CONTENT_TYPE_HTML),
            {'body_html': "No can do"}
        )
        mock_render.side_effect = None
        self.assertEqual(template._validate_body(EmailTemplate.CONTENT_TYPE_HTML), {})
        mock_render.side_effect = Exception("Something else")
        self.assertRaises(Exception, template._validate_body)

    @mock.patch.object(EmailTemplate, '_validate_body')
    def test_clean(self, mock_body):
        template = EmailTemplate()
        template.clean()
        mock_body.return_value = {'body_text': 'Template not found'}
        self.assertRaises(ValidationError, template.clean)

    def test_render_subject(self):
        template = EmailTemplate(subject='Hello {{ first_name }}')
        subject = template.render_subject({'first_name': 'fråd'})
        self.assertEqual(subject, 'Hello fråd')

    def test_render_body(self):
        template = EmailTemplate(
            body_text='Hello {{ first_name }}',
            body_html='<h1>Hello {{ first_name }}</h1>'
        )
        context = {'first_name': 'fråd'}
        self.assertEqual(template.render_body(context), 'Hello fråd')
        self.assertEqual(
            template.render_body(context, content_type=EmailTemplate.CONTENT_TYPE_PLAIN),
            'Hello fråd'
        )
        self.assertEqual(
            template.render_body(context, content_type=EmailTemplate.CONTENT_TYPE_HTML),
            '<h1>Hello fråd</h1>'
        )
        self.assertRaises(
            AssertionError,
            template.render_body,
            context,
            content_type='foo'
        )

    def test_create_message(self):
        template = EmailTemplate(
            subject='Welcome message',
            body_text='Hello {{ first_name }}',
            body_html='<h1>Hello {{ first_name }}</h1>'
        )
        context = {'first_name': 'fråd'}
        message = template.create_message(context)
        self.assertIsInstance(message, EmailMultiAlternatives)
        self.assertEqual(message.subject, 'Welcome message')
        self.assertEqual(message.body, 'Hello fråd')
        self.assertEqual(
            message.alternatives,
            [('<h1>Hello fråd</h1>', EmailTemplate.CONTENT_TYPE_HTML)]
        )
        self.assertEqual(message.to, [])
        self.assertEqual(message.cc, [])
        self.assertEqual(message.bcc, [])
        self.assertEqual(message.from_email, settings.DEFAULT_FROM_EMAIL)
        self.assertEqual(message.reply_to, [settings.DEFAULT_FROM_EMAIL])

        message = template.create_message(
            context,
            to=['bruce@kung.fu'],
            cc=['fred@example.com'],
            bcc=['ginger@example.com'],
            from_email='Fred <fred@example.com>'
        )
        self.assertEqual(message.to, ['bruce@kung.fu'])
        self.assertEqual(message.cc, ['fred@example.com'])
        self.assertEqual(message.bcc, ['ginger@example.com'])
        self.assertEqual(message.from_email, 'Fred <fred@example.com>')
        # and so on - not going to test every property.

        # but we will check the three illegal kwargs
        self.assertRaises(AssertionError, template.create_message, {}, subject='foo')
        self.assertRaises(AssertionError, template.create_message, {}, body='foo')
        self.assertRaises(AssertionError, template.create_message, {}, alternatives='foo')

    def test_clone_template(self):
        template = EmailTemplate(
            name='Test template',
            language='en-us',
            version=0
        ).save()
        pk = template.pk
        clone = template.clone()
        template = EmailTemplate.objects.get(id=pk)
        self.assertEqual(clone.name, template.name)
        self.assertEqual(clone.language, template.language)
        self.assertEqual(clone.version, 1)
        self.assertNotEqual(clone.id, template.id)



from unittest import mock

from django.core.exceptions import ValidationError
from django.core.mail import EmailMultiAlternatives
from django.forms import Textarea
from django.http import HttpRequest
from django.test import TestCase

from ..forms import (
    EmailTestForm,
    JSONWidget,
    MultiEmailField,
    MultiEmailTemplateField,
)
from ..models import EmailTemplate


class JSONWidgetTests(TestCase):

    def test_format_value(self):
        widget = JSONWidget()
        self.assertEqual(widget.format_value(None), '{}')
        self.assertEqual(widget.format_value(''), '{}')
        self.assertEqual(widget.format_value('{"foo": true}'), '{\n    "foo": true\n}')
        self.assertRaises(TypeError, widget.format_value, {"foo": True})

    def test_render(self):
        widget = JSONWidget()
        textarea = Textarea()
        for val in [None, '', '{"foo": true}']:
            self.assertEqual(
                widget.render('test', val),
                textarea.render('test', widget.format_value(val), attrs=widget.DEFAULT_ATTRS)
            )


class MultiEmailFieldTests(TestCase):

    def test_to_python(self):
        form = MultiEmailField()
        self.assertEqual(form.to_python(None), [])
        self.assertEqual(form.to_python(''), [])
        self.assertEqual(form.to_python('fred@example.com'), ['fred@example.com'])
        self.assertEqual(
            form.to_python('fred@example.com , ginger@example.com'),
            ['fred@example.com', 'ginger@example.com']
        )
        self.assertEqual(
            form.to_python(['fred@example.com']),
            ['fred@example.com']
        )

    def test_validate(self):
        form = MultiEmailField()
        form.validate(['fred@example.com'])
        form.validate(form.to_python('fred@example.com, ginger@example.com'))
        # single email address fails validation - must be a list
        self.assertRaises(ValidationError, form.validate, 'fred@example.com')


class MultiEmailTemplateFieldTests(TestCase):

    @mock.patch.object(EmailTemplate.objects, 'filter')
    def test_to_python(self, mock_filter):
        form = MultiEmailTemplateField()
        self.assertEqual(list(form.to_python(None)), list(EmailTemplate.objects.none()))
        self.assertEqual(list(form.to_python('')), list(EmailTemplate.objects.none()))
        qs = EmailTemplate.objects.none()
        self.assertEqual(form.to_python(qs), qs)
        form.to_python('1, 2')
        mock_filter.assert_called_once_with(pk__in=[1, 2])


class EmailTestFormTests(TestCase):

    def test_clean_context(self):
        form = EmailTestForm()
        form.cleaned_data = {'context': 'true'}
        self.assertEqual(form.clean_context(), True)
        form.cleaned_data['context'] = True
        self.assertRaises(ValidationError, form.clean_context)

    def test__create_message(self):
        form = EmailTestForm()
        form.cleaned_data = {
            'context': {'foo': 'bar'},
            'to': ['fred@example.com'],
            'cc': [],
            'bcc': [],
            'from_email': 'donotreply@example.com'
        }
        template = EmailTemplate()
        email = form._create_message(template)
        self.assertEqual(email.from_email, 'donotreply@example.com')
        self.assertEqual(email.to, ['fred@example.com'])
        self.assertEqual(email.cc, [])
        self.assertEqual(email.bcc, [])

    @mock.patch('appmail.forms.messages')
    @mock.patch.object(EmailMultiAlternatives, 'send')
    def test_send_emails(self, mock_send, mock_messages):
        template = EmailTemplate()
        form = EmailTestForm()
        form.cleaned_data = {
            'context': {'foo': 'bar'},
            'to': ['fred@example.com'],
            'cc': [],
            'bcc': [],
            'from_email': 'donotreply@example.com',
            'templates': [template]
        }
        request = HttpRequest()
        form.send_emails(request)
        mock_send.assert_called_once()
        mock_messages.success.assert_called_once()

        # test email failure
        mock_send.side_effect = Exception()
        form.send_emails(request)
        mock_messages.error.assert_called_once()
